/*
Copyright (c) 2010 Alberto Fajardo

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

The tweaks and modifications for Java:
Copyright (c) 2010, Werner Dittmann. 

The same permissions granted.
*/

package org.bouncycastle.crypto.engines;
/*
 * Use the following commad (on Linux for example) to generate the real
 * Java file:
 *    gcc -E Threefish1024.java.c | sed -e '/^#/d' > Threefish1024.java
 *
 * The mix.h contains Mix and UnMix macros that replace the Mix/UnMix calls
 * below with the real code. This produces unrolled loops. This technique
 * saves a lot of error prone editing.
 * sed just removes some lines that gcc -E inserted.
 */
#include "mix.h"

public class Threefish1024 extends ThreefishCipher {
        private final int CIPHER_SIZE = 1024;
        private final int CIPHER_QWORDS = CIPHER_SIZE / 64;
        private final int EXPANDED_KEY_SIZE = CIPHER_QWORDS + 1;

        public Threefish1024()
        {
            // Create the expanded key array
            ExpandedKey = new long[EXPANDED_KEY_SIZE];
            ExpandedKey[EXPANDED_KEY_SIZE - 1] = KeyScheduleConst;
        }

        public void Encrypt(long[] input, long[] output)
        {
            // Cache the block, key, and tweak
            long b0 = input[0], b1 = input[1],
                  b2 = input[2], b3 = input[3],
                  b4 = input[4], b5 = input[5],
                  b6 = input[6], b7 = input[7],
                  b8 = input[8], b9 = input[9],
                  b10 = input[10], b11 = input[11],
                  b12 = input[12], b13 = input[13],
                  b14 = input[14], b15 = input[15];
            long k0 = ExpandedKey[0], k1 = ExpandedKey[1],
                  k2 = ExpandedKey[2], k3 = ExpandedKey[3],
                  k4 = ExpandedKey[4], k5 = ExpandedKey[5],
                  k6 = ExpandedKey[6], k7 = ExpandedKey[7],
                  k8 = ExpandedKey[8], k9 = ExpandedKey[9],
                  k10 = ExpandedKey[10], k11 = ExpandedKey[11],
                  k12 = ExpandedKey[12], k13 = ExpandedKey[13],
                  k14 = ExpandedKey[14], k15 = ExpandedKey[15],
                  k16 = ExpandedKey[16];
            long t0 = ExpandedTweak[0], t1 = ExpandedTweak[1],
                  t2 = ExpandedTweak[2];


            Mix5( b0,  b1, 24, k0, k1);
            Mix5( b2,  b3, 13, k2, k3);
            Mix5( b4,  b5, 8, k4, k5);
            Mix5( b6,  b7, 47, k6, k7);
            Mix5( b8,  b9, 8, k8, k9);
            Mix5( b10,  b11, 17, k10, k11);
            Mix5( b12,  b13, 22, k12, k13 + t0);
            Mix5( b14,  b15, 37, k14 + t1, k15);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k1, k2);
            Mix5( b2,  b3, 9, k3, k4);
            Mix5( b4,  b5, 37, k5, k6);
            Mix5( b6,  b7, 31, k7, k8);
            Mix5( b8,  b9, 12, k9, k10);
            Mix5( b10,  b11, 47, k11, k12);
            Mix5( b12,  b13, 44, k13, k14 + t1);
            Mix5( b14,  b15, 30, k15 + t2, k16 + 1);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k2, k3);
            Mix5( b2,  b3, 13, k4, k5);
            Mix5( b4,  b5, 8, k6, k7);
            Mix5( b6,  b7, 47, k8, k9);
            Mix5( b8,  b9, 8, k10, k11);
            Mix5( b10,  b11, 17, k12, k13);
            Mix5( b12,  b13, 22, k14, k15 + t2);
            Mix5( b14,  b15, 37, k16 + t0, k0 + 2);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k3, k4);
            Mix5( b2,  b3, 9, k5, k6);
            Mix5( b4,  b5, 37, k7, k8);
            Mix5( b6,  b7, 31, k9, k10);
            Mix5( b8,  b9, 12, k11, k12);
            Mix5( b10,  b11, 47, k13, k14);
            Mix5( b12,  b13, 44, k15, k16 + t0);
            Mix5( b14,  b15, 30, k0 + t1, k1 + 3);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k4, k5);
            Mix5( b2,  b3, 13, k6, k7);
            Mix5( b4,  b5, 8, k8, k9);
            Mix5( b6,  b7, 47, k10, k11);
            Mix5( b8,  b9, 8, k12, k13);
            Mix5( b10,  b11, 17, k14, k15);
            Mix5( b12,  b13, 22, k16, k0 + t1);
            Mix5( b14,  b15, 37, k1 + t2, k2 + 4);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k5, k6);
            Mix5( b2,  b3, 9, k7, k8);
            Mix5( b4,  b5, 37, k9, k10);
            Mix5( b6,  b7, 31, k11, k12);
            Mix5( b8,  b9, 12, k13, k14);
            Mix5( b10,  b11, 47, k15, k16);
            Mix5( b12,  b13, 44, k0, k1 + t2);
            Mix5( b14,  b15, 30, k2 + t0, k3 + 5);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k6, k7);
            Mix5( b2,  b3, 13, k8, k9);
            Mix5( b4,  b5, 8, k10, k11);
            Mix5( b6,  b7, 47, k12, k13);
            Mix5( b8,  b9, 8, k14, k15);
            Mix5( b10,  b11, 17, k16, k0);
            Mix5( b12,  b13, 22, k1, k2 + t0);
            Mix5( b14,  b15, 37, k3 + t1, k4 + 6);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k7, k8);
            Mix5( b2,  b3, 9, k9, k10);
            Mix5( b4,  b5, 37, k11, k12);
            Mix5( b6,  b7, 31, k13, k14);
            Mix5( b8,  b9, 12, k15, k16);
            Mix5( b10,  b11, 47, k0, k1);
            Mix5( b12,  b13, 44, k2, k3 + t1);
            Mix5( b14,  b15, 30, k4 + t2, k5 + 7);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k8, k9);
            Mix5( b2,  b3, 13, k10, k11);
            Mix5( b4,  b5, 8, k12, k13);
            Mix5( b6,  b7, 47, k14, k15);
            Mix5( b8,  b9, 8, k16, k0);
            Mix5( b10,  b11, 17, k1, k2);
            Mix5( b12,  b13, 22, k3, k4 + t2);
            Mix5( b14,  b15, 37, k5 + t0, k6 + 8);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k9, k10);
            Mix5( b2,  b3, 9, k11, k12);
            Mix5( b4,  b5, 37, k13, k14);
            Mix5( b6,  b7, 31, k15, k16);
            Mix5( b8,  b9, 12, k0, k1);
            Mix5( b10,  b11, 47, k2, k3);
            Mix5( b12,  b13, 44, k4, k5 + t0);
            Mix5( b14,  b15, 30, k6 + t1, k7 + 9);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k10, k11);
            Mix5( b2,  b3, 13, k12, k13);
            Mix5( b4,  b5, 8, k14, k15);
            Mix5( b6,  b7, 47, k16, k0);
            Mix5( b8,  b9, 8, k1, k2);
            Mix5( b10,  b11, 17, k3, k4);
            Mix5( b12,  b13, 22, k5, k6 + t1);
            Mix5( b14,  b15, 37, k7 + t2, k8 + 10);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k11, k12);
            Mix5( b2,  b3, 9, k13, k14);
            Mix5( b4,  b5, 37, k15, k16);
            Mix5( b6,  b7, 31, k0, k1);
            Mix5( b8,  b9, 12, k2, k3);
            Mix5( b10,  b11, 47, k4, k5);
            Mix5( b12,  b13, 44, k6, k7 + t2);
            Mix5( b14,  b15, 30, k8 + t0, k9 + 11);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k12, k13);
            Mix5( b2,  b3, 13, k14, k15);
            Mix5( b4,  b5, 8, k16, k0);
            Mix5( b6,  b7, 47, k1, k2);
            Mix5( b8,  b9, 8, k3, k4);
            Mix5( b10,  b11, 17, k5, k6);
            Mix5( b12,  b13, 22, k7, k8 + t0);
            Mix5( b14,  b15, 37, k9 + t1, k10 + 12);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k13, k14);
            Mix5( b2,  b3, 9, k15, k16);
            Mix5( b4,  b5, 37, k0, k1);
            Mix5( b6,  b7, 31, k2, k3);
            Mix5( b8,  b9, 12, k4, k5);
            Mix5( b10,  b11, 47, k6, k7);
            Mix5( b12,  b13, 44, k8, k9 + t1);
            Mix5( b14,  b15, 30, k10 + t2, k11 + 13);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k14, k15);
            Mix5( b2,  b3, 13, k16, k0);
            Mix5( b4,  b5, 8, k1, k2);
            Mix5( b6,  b7, 47, k3, k4);
            Mix5( b8,  b9, 8, k5, k6);
            Mix5( b10,  b11, 17, k7, k8);
            Mix5( b12,  b13, 22, k9, k10 + t2);
            Mix5( b14,  b15, 37, k11 + t0, k12 + 14);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k15, k16);
            Mix5( b2,  b3, 9, k0, k1);
            Mix5( b4,  b5, 37, k2, k3);
            Mix5( b6,  b7, 31, k4, k5);
            Mix5( b8,  b9, 12, k6, k7);
            Mix5( b10,  b11, 47, k8, k9);
            Mix5( b12,  b13, 44, k10, k11 + t0);
            Mix5( b14,  b15, 30, k12 + t1, k13 + 15);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k16, k0);
            Mix5( b2,  b3, 13, k1, k2);
            Mix5( b4,  b5, 8, k3, k4);
            Mix5( b6,  b7, 47, k5, k6);
            Mix5( b8,  b9, 8, k7, k8);
            Mix5( b10,  b11, 17, k9, k10);
            Mix5( b12,  b13, 22, k11, k12 + t1);
            Mix5( b14,  b15, 37, k13 + t2, k14 + 16);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k0, k1);
            Mix5( b2,  b3, 9, k2, k3);
            Mix5( b4,  b5, 37, k4, k5);
            Mix5( b6,  b7, 31, k6, k7);
            Mix5( b8,  b9, 12, k8, k9);
            Mix5( b10,  b11, 47, k10, k11);
            Mix5( b12,  b13, 44, k12, k13 + t2);
            Mix5( b14,  b15, 30, k14 + t0, k15 + 17);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);
            Mix5( b0,  b1, 24, k1, k2);
            Mix5( b2,  b3, 13, k3, k4);
            Mix5( b4,  b5, 8, k5, k6);
            Mix5( b6,  b7, 47, k7, k8);
            Mix5( b8,  b9, 8, k9, k10);
            Mix5( b10,  b11, 17, k11, k12);
            Mix5( b12,  b13, 22, k13, k14 + t0);
            Mix5( b14,  b15, 37, k15 + t1, k16 + 18);
            Mix( b0,  b9, 38);
            Mix( b2,  b13, 19);
            Mix( b6,  b11, 10);
            Mix( b4,  b15, 55);
            Mix( b10,  b7, 49);
            Mix( b12,  b3, 18);
            Mix( b14,  b5, 23);
            Mix( b8,  b1, 52);
            Mix( b0,  b7, 33);
            Mix( b2,  b5, 4);
            Mix( b4,  b3, 51);
            Mix( b6,  b1, 13);
            Mix( b12,  b15, 34);
            Mix( b14,  b13, 41);
            Mix( b8,  b11, 59);
            Mix( b10,  b9, 17);
            Mix( b0,  b15, 5);
            Mix( b2,  b11, 20);
            Mix( b6,  b13, 48);
            Mix( b4,  b9, 41);
            Mix( b14,  b1, 47);
            Mix( b8,  b5, 28);
            Mix( b10,  b3, 16);
            Mix( b12,  b7, 25);
            Mix5( b0,  b1, 41, k2, k3);
            Mix5( b2,  b3, 9, k4, k5);
            Mix5( b4,  b5, 37, k6, k7);
            Mix5( b6,  b7, 31, k8, k9);
            Mix5( b8,  b9, 12, k10, k11);
            Mix5( b10,  b11, 47, k12, k13);
            Mix5( b12,  b13, 44, k14, k15 + t1);
            Mix5( b14,  b15, 30, k16 + t2, k0 + 19);
            Mix( b0,  b9, 16);
            Mix( b2,  b13, 34);
            Mix( b6,  b11, 56);
            Mix( b4,  b15, 51);
            Mix( b10,  b7, 4);
            Mix( b12,  b3, 53);
            Mix( b14,  b5, 42);
            Mix( b8,  b1, 41);
            Mix( b0,  b7, 31);
            Mix( b2,  b5, 44);
            Mix( b4,  b3, 47);
            Mix( b6,  b1, 46);
            Mix( b12,  b15, 19);
            Mix( b14,  b13, 42);
            Mix( b8,  b11, 44);
            Mix( b10,  b9, 25);
            Mix( b0,  b15, 9);
            Mix( b2,  b11, 48);
            Mix( b6,  b13, 35);
            Mix( b4,  b9, 52);
            Mix( b14,  b1, 23);
            Mix( b8,  b5, 31);
            Mix( b10,  b3, 37);
            Mix( b12,  b7, 20);

            // Final key schedule
            output[0] = b0 + k3;
            output[1] = b1 + k4;
            output[2] = b2 + k5;
            output[3] = b3 + k6;
            output[4] = b4 + k7;
            output[5] = b5 + k8;
            output[6] = b6 + k9;
            output[7] = b7 + k10;
            output[8] = b8 + k11;
            output[9] = b9 + k12;
            output[10] = b10 + k13;
            output[11] = b11 + k14;
            output[12] = b12 + k15;
            output[13] = b13 + k16 + t2;
            output[14] = b14 + k0 + t0;
            output[15] = b15 + k1 + 20;
        }

        public void Decrypt(long[] input, long[] output)
        {
            // Cache the block, key, and tweak
            long b0 = input[0], b1 = input[1],
                  b2 = input[2], b3 = input[3],
                  b4 = input[4], b5 = input[5],
                  b6 = input[6], b7 = input[7],
                  b8 = input[8], b9 = input[9],
                  b10 = input[10], b11 = input[11],
                  b12 = input[12], b13 = input[13],
                  b14 = input[14], b15 = input[15];
            long k0 = ExpandedKey[0], k1 = ExpandedKey[1],
                  k2 = ExpandedKey[2], k3 = ExpandedKey[3],
                  k4 = ExpandedKey[4], k5 = ExpandedKey[5],
                  k6 = ExpandedKey[6], k7 = ExpandedKey[7],
                  k8 = ExpandedKey[8], k9 = ExpandedKey[9],
                  k10 = ExpandedKey[10], k11 = ExpandedKey[11],
                  k12 = ExpandedKey[12], k13 = ExpandedKey[13],
                  k14 = ExpandedKey[14], k15 = ExpandedKey[15],
                  k16 = ExpandedKey[16];
            long t0 = ExpandedTweak[0], t1 = ExpandedTweak[1],
                  t2 = ExpandedTweak[2];
	    long tmp;

            b0 -= k3;
            b1 -= k4;
            b2 -= k5;
            b3 -= k6;
            b4 -= k7;
            b5 -= k8;
            b6 -= k9;
            b7 -= k10;
            b8 -= k11;
            b9 -= k12;
            b10 -= k13;
            b11 -= k14;
            b12 -= k15;
            b13 -= k16 + t2;
            b14 -= k0 + t0;
            b15 -= k1 + 20;
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k16 + t2, k0 + 19);
            UnMix5( b12,  b13, 44, k14, k15 + t1);
            UnMix5( b10,  b11, 47, k12, k13);
            UnMix5( b8,  b9, 12, k10, k11);
            UnMix5( b6,  b7, 31, k8, k9);
            UnMix5( b4,  b5, 37, k6, k7);
            UnMix5( b2,  b3, 9, k4, k5);
            UnMix5( b0,  b1, 41, k2, k3);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k15 + t1, k16 + 18);
            UnMix5( b12,  b13, 22, k13, k14 + t0);
            UnMix5( b10,  b11, 17, k11, k12);
            UnMix5( b8,  b9, 8, k9, k10);
            UnMix5( b6,  b7, 47, k7, k8);
            UnMix5( b4,  b5, 8, k5, k6);
            UnMix5( b2,  b3, 13, k3, k4);
            UnMix5( b0,  b1, 24, k1, k2);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k14 + t0, k15 + 17);
            UnMix5( b12,  b13, 44, k12, k13 + t2);
            UnMix5( b10,  b11, 47, k10, k11);
            UnMix5( b8,  b9, 12, k8, k9);
            UnMix5( b6,  b7, 31, k6, k7);
            UnMix5( b4,  b5, 37, k4, k5);
            UnMix5( b2,  b3, 9, k2, k3);
            UnMix5( b0,  b1, 41, k0, k1);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k13 + t2, k14 + 16);
            UnMix5( b12,  b13, 22, k11, k12 + t1);
            UnMix5( b10,  b11, 17, k9, k10);
            UnMix5( b8,  b9, 8, k7, k8);
            UnMix5( b6,  b7, 47, k5, k6);
            UnMix5( b4,  b5, 8, k3, k4);
            UnMix5( b2,  b3, 13, k1, k2);
            UnMix5( b0,  b1, 24, k16, k0);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k12 + t1, k13 + 15);
            UnMix5( b12,  b13, 44, k10, k11 + t0);
            UnMix5( b10,  b11, 47, k8, k9);
            UnMix5( b8,  b9, 12, k6, k7);
            UnMix5( b6,  b7, 31, k4, k5);
            UnMix5( b4,  b5, 37, k2, k3);
            UnMix5( b2,  b3, 9, k0, k1);
            UnMix5( b0,  b1, 41, k15, k16);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k11 + t0, k12 + 14);
            UnMix5( b12,  b13, 22, k9, k10 + t2);
            UnMix5( b10,  b11, 17, k7, k8);
            UnMix5( b8,  b9, 8, k5, k6);
            UnMix5( b6,  b7, 47, k3, k4);
            UnMix5( b4,  b5, 8, k1, k2);
            UnMix5( b2,  b3, 13, k16, k0);
            UnMix5( b0,  b1, 24, k14, k15);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k10 + t2, k11 + 13);
            UnMix5( b12,  b13, 44, k8, k9 + t1);
            UnMix5( b10,  b11, 47, k6, k7);
            UnMix5( b8,  b9, 12, k4, k5);
            UnMix5( b6,  b7, 31, k2, k3);
            UnMix5( b4,  b5, 37, k0, k1);
            UnMix5( b2,  b3, 9, k15, k16);
            UnMix5( b0,  b1, 41, k13, k14);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k9 + t1, k10 + 12);
            UnMix5( b12,  b13, 22, k7, k8 + t0);
            UnMix5( b10,  b11, 17, k5, k6);
            UnMix5( b8,  b9, 8, k3, k4);
            UnMix5( b6,  b7, 47, k1, k2);
            UnMix5( b4,  b5, 8, k16, k0);
            UnMix5( b2,  b3, 13, k14, k15);
            UnMix5( b0,  b1, 24, k12, k13);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k8 + t0, k9 + 11);
            UnMix5( b12,  b13, 44, k6, k7 + t2);
            UnMix5( b10,  b11, 47, k4, k5);
            UnMix5( b8,  b9, 12, k2, k3);
            UnMix5( b6,  b7, 31, k0, k1);
            UnMix5( b4,  b5, 37, k15, k16);
            UnMix5( b2,  b3, 9, k13, k14);
            UnMix5( b0,  b1, 41, k11, k12);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k7 + t2, k8 + 10);
            UnMix5( b12,  b13, 22, k5, k6 + t1);
            UnMix5( b10,  b11, 17, k3, k4);
            UnMix5( b8,  b9, 8, k1, k2);
            UnMix5( b6,  b7, 47, k16, k0);
            UnMix5( b4,  b5, 8, k14, k15);
            UnMix5( b2,  b3, 13, k12, k13);
            UnMix5( b0,  b1, 24, k10, k11);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k6 + t1, k7 + 9);
            UnMix5( b12,  b13, 44, k4, k5 + t0);
            UnMix5( b10,  b11, 47, k2, k3);
            UnMix5( b8,  b9, 12, k0, k1);
            UnMix5( b6,  b7, 31, k15, k16);
            UnMix5( b4,  b5, 37, k13, k14);
            UnMix5( b2,  b3, 9, k11, k12);
            UnMix5( b0,  b1, 41, k9, k10);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k5 + t0, k6 + 8);
            UnMix5( b12,  b13, 22, k3, k4 + t2);
            UnMix5( b10,  b11, 17, k1, k2);
            UnMix5( b8,  b9, 8, k16, k0);
            UnMix5( b6,  b7, 47, k14, k15);
            UnMix5( b4,  b5, 8, k12, k13);
            UnMix5( b2,  b3, 13, k10, k11);
            UnMix5( b0,  b1, 24, k8, k9);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k4 + t2, k5 + 7);
            UnMix5( b12,  b13, 44, k2, k3 + t1);
            UnMix5( b10,  b11, 47, k0, k1);
            UnMix5( b8,  b9, 12, k15, k16);
            UnMix5( b6,  b7, 31, k13, k14);
            UnMix5( b4,  b5, 37, k11, k12);
            UnMix5( b2,  b3, 9, k9, k10);
            UnMix5( b0,  b1, 41, k7, k8);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k3 + t1, k4 + 6);
            UnMix5( b12,  b13, 22, k1, k2 + t0);
            UnMix5( b10,  b11, 17, k16, k0);
            UnMix5( b8,  b9, 8, k14, k15);
            UnMix5( b6,  b7, 47, k12, k13);
            UnMix5( b4,  b5, 8, k10, k11);
            UnMix5( b2,  b3, 13, k8, k9);
            UnMix5( b0,  b1, 24, k6, k7);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k2 + t0, k3 + 5);
            UnMix5( b12,  b13, 44, k0, k1 + t2);
            UnMix5( b10,  b11, 47, k15, k16);
            UnMix5( b8,  b9, 12, k13, k14);
            UnMix5( b6,  b7, 31, k11, k12);
            UnMix5( b4,  b5, 37, k9, k10);
            UnMix5( b2,  b3, 9, k7, k8);
            UnMix5( b0,  b1, 41, k5, k6);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k1 + t2, k2 + 4);
            UnMix5( b12,  b13, 22, k16, k0 + t1);
            UnMix5( b10,  b11, 17, k14, k15);
            UnMix5( b8,  b9, 8, k12, k13);
            UnMix5( b6,  b7, 47, k10, k11);
            UnMix5( b4,  b5, 8, k8, k9);
            UnMix5( b2,  b3, 13, k6, k7);
            UnMix5( b0,  b1, 24, k4, k5);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k0 + t1, k1 + 3);
            UnMix5( b12,  b13, 44, k15, k16 + t0);
            UnMix5( b10,  b11, 47, k13, k14);
            UnMix5( b8,  b9, 12, k11, k12);
            UnMix5( b6,  b7, 31, k9, k10);
            UnMix5( b4,  b5, 37, k7, k8);
            UnMix5( b2,  b3, 9, k5, k6);
            UnMix5( b0,  b1, 41, k3, k4);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k16 + t0, k0 + 2);
            UnMix5( b12,  b13, 22, k14, k15 + t2);
            UnMix5( b10,  b11, 17, k12, k13);
            UnMix5( b8,  b9, 8, k10, k11);
            UnMix5( b6,  b7, 47, k8, k9);
            UnMix5( b4,  b5, 8, k6, k7);
            UnMix5( b2,  b3, 13, k4, k5);
            UnMix5( b0,  b1, 24, k2, k3);
            UnMix( b12,  b7, 20);
            UnMix( b10,  b3, 37);
            UnMix( b8,  b5, 31);
            UnMix( b14,  b1, 23);
            UnMix( b4,  b9, 52);
            UnMix( b6,  b13, 35);
            UnMix( b2,  b11, 48);
            UnMix( b0,  b15, 9);
            UnMix( b10,  b9, 25);
            UnMix( b8,  b11, 44);
            UnMix( b14,  b13, 42);
            UnMix( b12,  b15, 19);
            UnMix( b6,  b1, 46);
            UnMix( b4,  b3, 47);
            UnMix( b2,  b5, 44);
            UnMix( b0,  b7, 31);
            UnMix( b8,  b1, 41);
            UnMix( b14,  b5, 42);
            UnMix( b12,  b3, 53);
            UnMix( b10,  b7, 4);
            UnMix( b4,  b15, 51);
            UnMix( b6,  b11, 56);
            UnMix( b2,  b13, 34);
            UnMix( b0,  b9, 16);
            UnMix5( b14,  b15, 30, k15 + t2, k16 + 1);
            UnMix5( b12,  b13, 44, k13, k14 + t1);
            UnMix5( b10,  b11, 47, k11, k12);
            UnMix5( b8,  b9, 12, k9, k10);
            UnMix5( b6,  b7, 31, k7, k8);
            UnMix5( b4,  b5, 37, k5, k6);
            UnMix5( b2,  b3, 9, k3, k4);
            UnMix5( b0,  b1, 41, k1, k2);
            UnMix( b12,  b7, 25);
            UnMix( b10,  b3, 16);
            UnMix( b8,  b5, 28);
            UnMix( b14,  b1, 47);
            UnMix( b4,  b9, 41);
            UnMix( b6,  b13, 48);
            UnMix( b2,  b11, 20);
            UnMix( b0,  b15, 5);
            UnMix( b10,  b9, 17);
            UnMix( b8,  b11, 59);
            UnMix( b14,  b13, 41);
            UnMix( b12,  b15, 34);
            UnMix( b6,  b1, 13);
            UnMix( b4,  b3, 51);
            UnMix( b2,  b5, 4);
            UnMix( b0,  b7, 33);
            UnMix( b8,  b1, 52);
            UnMix( b14,  b5, 23);
            UnMix( b12,  b3, 18);
            UnMix( b10,  b7, 49);
            UnMix( b4,  b15, 55);
            UnMix( b6,  b11, 10);
            UnMix( b2,  b13, 19);
            UnMix( b0,  b9, 38);
            UnMix5( b14,  b15, 37, k14 + t1, k15);
            UnMix5( b12,  b13, 22, k12, k13 + t0);
            UnMix5( b10,  b11, 17, k10, k11);
            UnMix5( b8,  b9, 8, k8, k9);
            UnMix5( b6,  b7, 47, k6, k7);
            UnMix5( b4,  b5, 8, k4, k5);
            UnMix5( b2,  b3, 13, k2, k3);
            UnMix5( b0,  b1, 24, k0, k1);

            output[15] = b15;
            output[14] = b14;
            output[13] = b13;
            output[12] = b12;
            output[11] = b11;
            output[10] = b10;
            output[9] = b9;
            output[8] = b8;
            output[7] = b7;
            output[6] = b6;
            output[5] = b5;
            output[4] = b4;
            output[3] = b3;
            output[2] = b2;
            output[1] = b1;
            output[0] = b0;
        }
}
